/* 
Basic use of the web server

var server = new HTTPD()  //creates a web server instance
server.run(8080,true)     //runs the server on port 8080, and tries to open a browser to view it

Advanced use

var server = new HTTPD()
server.hosts = ['localhost']     //restrict access
server.handlers["jsp"] = execJSP //enable server-parsed JavaScript pages 
                                 //you can add special handlers for other types, too



*/

/* To do: adjust Expires and If-modified-since to return 304 Not Modified */

function HTTPD()
{
 this.http = null;
 this.running = true; //set this.running = false to exit the server
 this.types = {'html': "text/html; charset=utf-8", 
              'htm': "text/html", 
              'png': "image/png", 
              'gif': "image/gif", 
              'jpeg': "image/jpeg", 
              'jpg': "image/jpeg", 
              'txt': "text/plain",
              'css': "text/css",
              'svg': "image/svg+xml",
              'js': "application/x-javascript",
              '*': "application/x-unknown"}
 this.handlers = {'jsp': HTTPD.execJSP, 'cgi': HTTPD.execCGI}
 this.functions = {}
 /*this.functions["quit"] = function(client, data, httpd)
  {
    httpd.running = false;
    client.writeln("Content-type: text/html");
    client.writeln()
    client.close()
  }
 */ 
 this.hosts = [system.name,'localhost']
 this.index = ''
 this.forbidden = ['httpd.js']
}

HTTPD.prototype.home = function(client,data)
{
 client.writeln("Content-type: text/html");
 client.writeln();

 client.writeln("<html><body><p>Hello, world!</p>")

 client.writeln("HTTP headers\n<pre>"); 
 if (client.header)
 client.writeln(client.header.toString())
 client.writeln("</pre>")

 client.writeln("Form fields\n<pre>");
 if (data)
 client.writeln(data.toString())
 client.writeln("</pre>")

 client.writeln("</body></html>");
}

HTTPD.prototype.sendOK = function(client)
{
 client.writeln("HTTP/1.1 200 OK");
 client.writeln("Client: close");
 client.writeln("Date: ", client.startTime.toUTCString());
 client.writeln("Expires: ", 0);
 client.writeln("Server: JSDB/"+system.version);
}

HTTPD.prototype.sendERROR = function(client)
{
 client.writeln("HTTP/1.1 404 NOT FOUND");
 client.writeln("Client: close");
 client.writeln("Date: ", client.startTime.toUTCString());
 client.writeln("Expires: ", 0);
 client.writeln("Server: JSDB/"+system.version);
 client.writeln("Content-type: text/html\n");
 client.writeln("<H2>HTTP/1.1 404 Not Found</H2>");
 client.writeln("<br>URL:",client.uri);
 client.writeln("<br><a href=/>Home</a>");
 client.close();
}

HTTPD.prototype.run = function(port, browser)
{
 this.start(port || 8080,browser)
 while (this.running) // && !system.kbhit()
   if (!this.step())
     system.sleep(100)
 this.finish()
}

HTTPD.prototype.stop = function()
{
 this.running = false
}

HTTPD.prototype.start = function(port, browser)
{
 if (!port) port = 8080
 if (this.http != null) return;

 this.http = null;
 for (; this.http == null && port < 8180; port++)
  {
  try {
   this.http = new Server(port);
   } catch(err)
   {
    writeln('Port ', port,' appears to be in use');
   }
  }

 if (this.hosts.indexOf(this.http.hostName) == -1)
   this.hosts.push(this.http.hostName)
 
 writeln("Server started on port " + this.http.port);
 if (browser)
 {
  let url = 'http://'+this.http.toString() + '/' + (typeof browser == "string" ? browser : "")
  system.browse(url)
  writeln("Browsing ",url);
 }

 this.running = true
// while (this.running && !system.kbhit())
}

HTTPD.prototype.step = function()
 {
   system.gc()
   if (!this.http.anyoneWaiting)
    {
     return false;
    }

   let client = this.http.accept();
   if (client == null)
    return true;
      
   if (this.hosts.length && this.hosts.indexOf(client.name) == -1)
   {
    if (this.debug) writeln('Rejected ',client.name)
    client.close()
    return true;
   }
   if (this.debug) writeln('Connected ',client.name)
   return this.handleRequest(client,client.readLine())
}

HTTPD.prototype.handleRequest = function(client,request)
{
   request = request.split(/\s+/); //GET /index.html HTTP/1.1
   if (this.debug) writeln(request)

   client.method = request[0];
   client.uri = request[1];
   if (client.uri == null || client.uri == '') client.uri = '/';
   client.version = request[2];
   //this.page() should give enough time for the header packet to arrive
   
   if (!client.canRead) system.sleep(50)
   client.startTime = new Date();
   client.header = new Record;
   client.readMIME(client.header);
   
   client.page = decodeURI(client.uri.substr(1));
   client.query = ''
   request = client.uri.match(/\/?([^?]*)\?(.*)/);
   if (request != null)
    {
     client.page = request[1];
     client.query = request[2];
    }

   let rawData = null
   client.data = {}
   if (client.method == "GET" && client.query)
      rawData = client.query
   else if (client.method = "POST" && client.header && client.header.get('Content-type')  == 'application/x-www-form-urlencoded')
      rawData = client.read(client.header.get('Content-length'))

   if (rawData)
    {
      let source = rawData.split(/\&/g)
      for (let x=0; x< source.length; x++)
      {
       let [name,value] = source[x].split('=')
       client.data[decodeURL(name)] = value ? decodeURL(value) : ""
      }
    }
/*
   if (client.method == "GET" && client.query)
    client.data = new Record(client.query,'&');
   else if (client.method = "POST" && client.header.get('Content-type') == 'application/x-www-form-urlencoded')
    client.data = new Record(client.read(client.header.get('Content-length')),'&');

   if (client.data)
   {
    for(x=0; x<client.data.length; x++)
      client.data.set(x, decodeURL(client.data.value(x)));
   }
*/
   if (client.page == '')
   {
    if (typeof this.index == "string")
    {
     client.page = this.index
    }
    else if (typeof this.index == "function")
    {
      this.sendOK(client);      
      let done = this.index(client,client.data,this)

      if (typeof done == "string")
      {
        client.page = done;
      }
      else
      {
        client.close();
        return true;
      }
    }
   }
                                     
   if (this.functions[client.page]) //special function handler
   {
      let handler = this.functions[client.page]
      if (this.debug) writeln('executing ',handler.name,' for ',client.page)
      this.sendOK(client);      
      let done = handler(client,client.data,this)
      
      if (typeof done == "string")
      {
        client.page = done;
      }
      else
      {
        client.close();
        return true;
      }
   }

   if (client.page == 'test')
   {
    try 
    {
     this.sendOK(client);
     this.home(client);
    } 
    catch(err)
    {
     writeln("Error: ",err);
     client.writeln("Error: ",err);
    }
    client.close();
    return true;
   }
   
   // filter the file names. No URLs, wildcards, or path changes
   if (client.page.search(/(\\|\/|\*)/) != -1)
   {
    client.close()
    return true;
   }

   if (client.page && system.exists(client.page))
   {
     let page = client.page
     //leaves room for a redirection table later
     if (this.debug) writeln(client.name,'\t',page,'\t',client.query,'\t',new Date())
     if (this.debug) writeln(client.header)
     if (this.forbidden.indexOf(client.page) != -1)
       this.sendERROR(client) //forbidded file
       
     let type = page.match(/\.(.+)$/)

     let handler = this.handlers[type[1]]
     if (handler)
     {
      this.sendOK(client);
      if (this.debug) writeln('executing ',handler.name,' for ',client.page)
      try
      {
       handler(this,client,page);
      }
    catch(err)
    {
     writeln("Error calling ",page,': ',err)   
     for (let [n,v] in err) writeln(n,': ',v)
      }
      client.close();
      return true;
     }
     
     if (type) type = this.types[type[1].toLowerCase()]
     if (!type) type = this.types['*']
     
     if (!type)
     {
      this.sendERROR(client) //forbidded file type
      writeln("Unknown MIME type for ",page)
      return true;
     }

     this.sendOK(client);
     
     try
     {
      let src = new Stream("file://" + page,"rb")
      client.writeln("Content-type: ",type);
      client.writeln("Content-length: ",src.size);
      client.writeln();
      client.append(src);
      src.close();
     }
     catch(err)
     {
     writeln("Error opening ",page,': ',x)   
     for (let [n,v] in x) writeln(n,': ',v)
     }
     client.close();
     return true;
   }

   if (!client.page)
   {
     this.sendOK(client);
     this.home(client);
     client.close()
     return true;
   }
   writeln("Can't find ",client.page)
   this.sendERROR(client)
   client.close();
   return true;
 }

HTTPD.prototype.finish = function()
{
 this.http.close();
}

HTTPD.execJSP = function(httpd, client, page)
{
if (httpd.debug) 
{
 writeln(page,': ',client.data.toSource())
}
try
 {  
  /*
  let header = "function include(page) {return this.execJSP(this.httpd,client,page);}\n" + 
               "function print() {this.print.apply(this,arguments);}\n"
  client.execJSP = execJSP
  */

  function include(page) {execJSP(httpd,client,page);}
  oldprint = print
  oldprintln = println
  print = function () {client.write.apply(client,arguments);}
  println = function () {client.writeln.apply(client,arguments);}
  httpData = function() {return [client.data, 
      {cookie: client.header['Cookie'], hostName: client.name}];}

  let src  = new Stream("file://" +page,'rt')
 
  if (src.read(2) == "#!")
   src.readLine()
  else
   src.rewind()
  client.writeln("Content-type: text/html")
  client.writeln()  
  
  let text = new Stream
  
  let block = 1
  while (!src.eof)
  { 
    src.readUntil("<%",client)
    if (src.eof) break
    text.clear()
    text.write('let [data, cookie, hostName] = httpData(); ')
    src.readUntil("%>",text)
    text.rewind()
    try 
    {
     run(text,page + ":" + (block++))
    }
    catch(err) 
    {
     writeln('Error in ',x.fileName,':',x.lineNumber,' ',x.message)
     throw(x)
    }
  }
  print = oldprint
  println = oldprintln
  delete data
  src.close()
 }
 catch(err)
 {
   httpd.sendERROR(client)
   writeln("Error in ",page,": ",err);
   for (let [n,v] in err) writeln(n,': ',v)
   client.writeln("<h1>Error in ",page,": ",err,"</h1><pre>");
   for (let [n,v] in err) client.writeln(n,': ',v)
   client.writeln("</pre>")
 }
 system.gc()
}

HTTPD.execCGI = function(httpd, client, page)
{
    let data = client.data
    if (httpd.debug) writeln(page,' ',data.toSource())
    let cgi = null
    try 
    {
      let cgi;
      if (system.exists('jsdb.exe')) cgi = new Stream('exec://jsdb.exe ' + page);  
      else if (system.exists('jsdb')) cgi = new Stream('exec://./jsdb ' + page);  
      else cgi = new Stream('exec://jsdb ' + page);  

      cgi.writeln(data.toSource())
      if (httpd.debug)
      {
       let result = cgi.readFile()
       write(result)
       client.write(result)
      }
      else
       client.append(cgi)
      if (httpd.debug) writeln('exited')
    } 
    catch(err)
    {
     writeln("Error in ",page,': ',err)   
     for (let [n,v] in err) writeln(n,': ',v)
     client.writeln("<h1>Error in ",page,": ",err,"</h1><pre>");
     for (let [n,v] in err) client.writeln(n,': ',v)
     client.writeln("</pre>")
    } 
    if (cgi) cgi.close()
}

